use actix_web::{post, web::Data, HttpRequest, HttpResponse};
use chrono::{Duration, Utc};
use ipnetwork::IpNetwork;
use redis::AsyncCommands;
use serde::Deserialize;
use std::{collections::HashSet, fmt::Write, net::IpAddr};

use crate::{
    ctx::Ctx,
    db::models::{Account, Ban},
    error::NekrochanError,
    qsform::QsForm,
    web::{
        actions::{get_posts_from_ids, ActionTemplate},
        tcx::{account_from_auth, TemplateCtx},
        template_response,
    },
};

#[derive(Deserialize)]
pub struct StaffPostActionsForm {
    #[serde(default)]
    pub posts: Vec<String>,
    #[serde(rename = "staff_remove_posts")]
    pub remove_posts: Option<String>,
    #[serde(rename = "staff_remove_files")]
    pub remove_files: Option<String>,
    #[serde(rename = "staff_toggle_spoiler")]
    pub toggle_spoiler: Option<String>,
    pub remove_by_ip_board: Option<String>,
    pub remove_by_ip_global: Option<String>,
    pub toggle_sticky: Option<String>,
    pub toggle_lock: Option<String>,
    pub remove_reports: Option<String>,
    pub ban_user: Option<String>,
    pub ban_reporters: Option<String>,
    pub global_ban: Option<String>,
    pub unappealable_ban: Option<String>,
    pub ban_reason: Option<String>,
    pub ban_duration: Option<u64>,
    pub ban_range: Option<String>,
    pub troll_user: Option<String>,
}

#[post("/actions/staff-post-actions")]
pub async fn staff_post_actions(
    ctx: Data<Ctx>,
    req: HttpRequest,
    QsForm(form): QsForm<StaffPostActionsForm>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;
    let account = account_from_auth(&ctx, &req).await?;
    let posts = get_posts_from_ids(&ctx, &form.posts).await;

    let mut response = String::new();

    let mut posts_removed = 0;
    let mut files_removed = 0;
    let mut spoilers_toggled = 0;
    let mut stickies_toggled = 0;
    let mut locks_toggled = 0;
    let mut reports_removed = 0;
    let mut bans_issued = 0;
    let mut users_trolled = 0;

    for post in &posts {
        if (form.remove_posts.is_some()
            || form.remove_files.is_some()
            || form.toggle_spoiler.is_some()
            || form.toggle_sticky.is_some()
            || form.toggle_lock.is_some())
            && !(account.perms().owner() || account.perms().manage_posts())
        {
            writeln!(
                &mut response,
                "[Chyba] Nemáš oprávnění spravovat příspěvky."
            )
            .ok();

            continue;
        }

        if form.remove_posts.is_some() {
            post.delete(&ctx).await?;
            posts_removed += 1;
        }

        if form.remove_files.is_some() {
            post.delete_files(&ctx).await?;
            files_removed += post.files.0.len();
        }

        if form.toggle_spoiler.is_some() {
            post.update_spoiler(&ctx).await?;
            spoilers_toggled += post.files.0.len();
        }

        if form.remove_by_ip_board.is_some() {
            let key = format!("by_ip:{}", post.ip);
            let ip_posts: Vec<String> = ctx.cache().zrange(key, 0, -1).await?;
            let board_ip_posts = ip_posts
                .into_iter()
                .filter(|p| p.starts_with(&format!("{}/", post.board)))
                .collect::<Vec<_>>();

            for post in get_posts_from_ids(&ctx, &board_ip_posts).await {
                post.delete(&ctx).await?;
                posts_removed += 1;
            }
        }

        if form.remove_by_ip_global.is_some() {
            let key = format!("by_ip:{}", post.ip);
            let ip_posts: Vec<String> = ctx.cache().zrange(key, 0, -1).await?;

            for post in get_posts_from_ids(&ctx, &ip_posts).await {
                post.delete(&ctx).await?;
                posts_removed += 1;
            }
        }

        if form.toggle_sticky.is_some() {
            post.update_sticky(&ctx).await?;
            stickies_toggled += 1;
        }

        if form.toggle_lock.is_some() {
            if post.thread.is_some() {
                writeln!(&mut response, "[Chyba] Odpověď nelze uzamknout.").ok();
            } else {
                post.update_lock(&ctx).await?;
                locks_toggled += 1;
            }
        }
    }

    for post in &posts {
        if form.remove_reports.is_some() {
            if !(tcx.perms.owner() || tcx.perms.reports()) {
                writeln!(&mut response, "[Chyba] Nemáš přístup k hlášením.").ok();
                continue;
            }

            post.delete_reports(&ctx).await?;
            reports_removed += post.reports.0.len();
        }
    }

    let mut already_banned = HashSet::new();

    for post in &posts {
        if let ((Some(_), _) | (_, Some(_)), reason, duration, Some(range)) = (
            (form.ban_user.clone(), form.ban_reporters.clone()),
            form.ban_reason.clone().unwrap_or_default(),
            form.ban_duration.unwrap_or_default(),
            form.ban_range.clone(),
        ) {
            if !(account.perms().owner() || account.perms().bans()) {
                writeln!(&mut response, "[Chyba] Nemáš oprávnění vydat ban.").ok();
                continue;
            }

            let mut ips_to_ban = HashSet::new();

            if form.ban_user.is_some() && !already_banned.contains(&post.ip) {
                ips_to_ban.insert(post.ip);
            }

            if form.ban_reporters.is_some() {
                if !(tcx.perms.owner() || tcx.perms.reports()) {
                    writeln!(&mut response, "[Chyba] Nemáš přístup k hlášením.").ok();
                    continue;
                }

                ips_to_ban.extend(
                    post.reports
                        .0
                        .iter()
                        .map(|r| r.reporter_ip)
                        .filter(|ip| !already_banned.contains(ip)),
                )
            }

            if ips_to_ban.is_empty() {
                continue;
            }

            for ip in &ips_to_ban {
                ban_ip(
                    &ctx,
                    &account,
                    &form,
                    *ip,
                    post.board.clone(),
                    reason.clone(),
                    duration,
                    &range,
                )
                .await?;
            }

            if form.ban_user.is_some() {
                let content_nomarkup = format!(
                    "{}\n\n##(UŽIVATEL BYL ZA TENTO PŘÍSPĚVEK ZABANOVÁN)##",
                    post.content_nomarkup
                );

                let content = format!(
                    "{}\n\n<span class=\"jannytext\">(UŽIVATEL BYL ZA TENTO PŘÍSPĚVEK ZABANOVÁN)</span>",
                    post.content
                );

                post.update_content(&ctx, content, content_nomarkup).await?;
            }

            bans_issued += ips_to_ban.len();
            already_banned.extend(ips_to_ban);
        }
    }

    for post in &posts {
        if form.troll_user.is_some() {
            if !(tcx.perms.owner() || tcx.perms.edit_posts()) {
                writeln!(
                    &mut response,
                    "[Chyba] Nemáš oprávnění upravovat příspěvky."
                )
                .ok();
                continue;
            }

            if !(tcx.perms.owner() || tcx.perms.view_ips()) {
                writeln!(
                    &mut response,
                    "[Chyba] Nemáš oprávnění zobrazovat IP adresy."
                )
                .ok();
                continue;
            }

            let content_nomarkup = format!("{}\n\n##({})##", post.content_nomarkup, post.ip);

            let content = format!(
                "{}\n\n<span class=\"jannytext\">({})</span>",
                post.content, post.ip
            );

            post.update_content(&ctx, content, content_nomarkup).await?;
            users_trolled += 1;
        }
    }

    if posts_removed != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Odstraněny příspěvky: {posts_removed}"
        )
        .ok();
    }

    if files_removed != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Odstraněny soubory: {files_removed}"
        )
        .ok();
    }

    if spoilers_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Přepnuty spoilery: {spoilers_toggled}"
        )
        .ok();
    }

    if stickies_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Připnuty/odepnuty příspěvky: {stickies_toggled}"
        )
        .ok();
    }

    if locks_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Zamčena/odemčena vlákna: {locks_toggled}"
        )
        .ok();
    }

    if reports_removed != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Odstraněna hlášení: {reports_removed}"
        )
        .ok();
    }

    if users_trolled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Vytroleni uživatelé: {users_trolled}"
        )
        .ok();
    }

    if bans_issued != 0 {
        writeln!(&mut response, "[Úspěch] Uděleny bany: {bans_issued}").ok();
    }

    let template = ActionTemplate { tcx, response };

    template_response(&template)
}

#[allow(clippy::too_many_arguments)]
async fn ban_ip(
    ctx: &Ctx,
    account: &Account,
    form: &StaffPostActionsForm,
    ip: IpAddr,
    board: String,
    reason: String,
    duration: u64,
    range: &str,
) -> Result<(), NekrochanError> {
    let account = account.username.clone();

    let board = if form.global_ban.is_none() {
        Some(board)
    } else {
        None
    };

    let prefix = if ip.is_ipv4() {
        match range {
            "lan" => 24,
            "isp" => 16,
            _ => 32,
        }
    } else {
        match range {
            "lan" => 48,
            "isp" => 24,
            _ => 128,
        }
    };

    let ip_range = IpNetwork::new(ip, prefix)?;
    let reason: String = reason.trim().into();

    if reason.is_empty() || reason.len() > 200 {
        return Err(NekrochanError::BanReasonFormatError);
    }

    let appealable = form.unappealable_ban.is_none();

    let expires = if duration == 0 {
        None
    } else {
        Some(Utc::now() + Duration::days(duration as i64))
    };

    Ban::create(ctx, account, board, ip_range, reason, appealable, expires).await?;

    Ok(())
}
