use actix_web::{
    get,
    web::{Data, Path, Query},
    HttpRequest, HttpResponse,
};
use askama::Template;
use serde::Deserialize;
use std::{collections::HashMap, net::IpAddr};

use crate::{
    ctx::Ctx,
    db::models::{Board, Post},
    error::NekrochanError,
    filters,
    web::{tcx::TemplateCtx, template_response},
};

#[derive(Deserialize)]
pub struct IpPostsQuery {
    page: i64,
}

#[derive(Template)]
#[template(path = "ip-posts.html")]
struct IpPostsTemplate {
    tcx: TemplateCtx,
    ip: IpAddr,
    boards: HashMap<String, Board>,
    posts: Vec<Post>,
    page: i64,
}

#[get("/ip-posts/{ip}")]
pub async fn ip_posts(
    ctx: Data<Ctx>,
    req: HttpRequest,
    path: Path<IpAddr>,
    query: Option<Query<IpPostsQuery>>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;

    if !(tcx.perms.owner() || tcx.perms.view_ips()) {
        return Err(NekrochanError::InsufficientPermissionError);
    }

    let ip = path.into_inner();
    let boards = Board::read_all_map(&ctx).await?;
    let page = query.map_or(1, |q| q.page);

    if page <= 0 {
        return Err(NekrochanError::InvalidPageError);
    }

    let posts = Post::read_ip_page(&ctx, ip, page).await?;

    let template = IpPostsTemplate {
        tcx,
        ip,
        boards,
        posts,
        page,
    };

    template_response(&template)
}
