use std::collections::HashMap;

use actix_web::{
    get,
    web::{Data, Query},
    HttpRequest, HttpResponse,
};
use askama::Template;
use serde::Deserialize;

use crate::{
    ctx::Ctx,
    db::models::{Board, Post},
    error::NekrochanError,
    filters,
    web::{tcx::TemplateCtx, template_response},
};

#[derive(Deserialize)]
pub struct BoardQuery {
    page: i64,
}

#[allow(dead_code)]
#[derive(Template)]
#[template(path = "staff/reports.html")]
struct ReportsTemplate {
    tcx: TemplateCtx,
    boards: HashMap<String, Board>,
    posts: Vec<Post>,
    page: i64,
}

#[get("/staff/reports")]
async fn reports(
    ctx: Data<Ctx>,
    req: HttpRequest,
    query: Option<Query<BoardQuery>>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;

    if !(tcx.perms.owner() || tcx.perms.reports()) {
        return Err(NekrochanError::InsufficientPermissionError);
    }

    let boards = Board::read_all_map(&ctx).await?;
    let page = query.map_or(1, |q| q.page);

    if page <= 0 {
        return Err(NekrochanError::InvalidPageError);
    }

    let posts = Post::read_reports_page(&ctx, page).await?;

    let template = ReportsTemplate {
        tcx,
        boards,
        posts,
        page,
    };

    template_response(&template)
}
